SUBROUTINE hinich_bispec(x,n,m,ndim,sx,z,B,isn)
!***********************************************************************
!  FORTRAN 90 subroutine to compute estimate of normalized bispectrum
!  according to Hinich's 1982 paper.
!  File: hinich.f90
!  Coded by: Nusrat Jahan
!
!  INPUT: 
!  x      =  A double precision vector of length n containing the
!            time series.
!  n      =  An integer scalar containing the length of the time series.
!  m      =  INT(SQRT(DBLE(n))) (divisor for bispectrum).
!  max_n  =  integer scalar containing the maximum number of rows
!            of z, f2hat in the calling program.
!
!  OUTPUT:
!  fhat   = A real double precision real vector containing the 
!           periodogram.
!  z      = A real double precision matrix of dimension m x m 
!           containing the square modulus of the estimated normalized
!           bispectrum.
!  B      = A double precision complex matrix of dimension m x m
!           containing the estimated bispectrum.
!
!  Reference:
!  Jahan, N. and Harvill, J.L. (2008).
!    Bispectral-based goodness-of-fit tests of Gaussianity
!    and linearity of stationary time series.
!    Communications in Statistics: Theory and Methods, 37(20), 3216--3227.
!    DOI: 10.1080/03610920802133319.
!*************************************************************************
  USE define_types
! USE ts_utilities
  IMPLICIT NONE
!
  INTEGER(i4b), INTENT(in) :: m, n, ndim
  INTEGER(i4b), INTENT(inout) :: isn
  REAL(dp), DIMENSION(n), INTENT(in) :: x
  REAL(dp), DIMENSION(m), INTENT(inout) :: sx
  REAL(dp), DIMENSION(ndim, m), INTENT(inout) :: z
  COMPLEX(dpc), DIMENSION(ndim, m), INTENT(inout) :: B
!
  INTEGER(i4b) :: j, k, jj, kk, nq, jpk
  INTEGER(i4b), DIMENSION(m) :: ix
  INTEGER(i4b), DIMENSION(m, m) :: i1, iq
  REAL(dp) :: dn, dm, wj, wk, sxj, sxk, sx1, sx2, sxjpk, div
  REAL(dp), DIMENSION(m+1) :: xx
  REAL(dp), DIMENSION(n) :: xr, xi
  COMPLEX(dpc) :: xj, xk
!
!  Get grid points of boxes:
!
!  First and last are 1/(2n) below 0 and above 0.5:
!
  dn = DBLE(n)
  dm = DBLE(m)
  xx(1)   = -1.0_dp/(2.0_dp*dn)
  xx(m+1) = 0.5_dp + (1.0_dp/(2.0_dp*dn))
!
!  The jth is average of the two natural frequencies "containing"
!  w = (j-1)/2m (if w is a natural frequency, then grid point is
!  average of that frequency and next largest one.
!
  DO j = 2, m
     jj    = dn*(j-1)/(2.0_dp*dm)
     xx(j) = DBLE(jj+jj+1)/(2.0_dp*dn)
  END DO
!
!  Now get fft of data:
!
  xr = x
  xi = 0.0_dp
!
  CALL dfft(xr,xi,n,n,n,isn)
  IF(isn .NE. 1) RETURN
!
!  Zero out zero frequency:
!
  xr(1) = 0.0_dp
  xi(1) = 0.0_dp
!
!  Zero out all arrays used to accumulate sums:
!
  sx = 0.0_dp
  ix = 0
  i1 = 0
  iq = 0
  z  = 0.0_dp
  B  = CMPLX(0.0_dp,0.0_dp)
!
!  nq is the number of natural frequencies:
!
  nq = (n/2) + 1
!
!   Loop over natural frequencies in two dimensions, accumulating
!   sums. The following variables are used:
!
!     j, k: the number of the horizontal and vertical 
!           natural frequencies
!     wj, wk: the corresponding frequencies
!     jj, kk: which column (row) of boxes the jth horizontal natural 
!             frequency (kth vertical natural frequency) belongs to
!     xj, xk: the corresponding fft values
!     sx(1),...,sx(m): accumulate stuff for autospectra estimate
!     ix(1),...,ix(m): accumulate number of values in average for sx
!     B(jj,kk) : accumulates bispectra estimate
!     i1(jj,kk): number of terms in B(jj,kk)
!
  jj = 1
  DO j = 1, nq
     xj = CMPLX(xr(j),xi(j))
     wj = (DBLE(j) - 1.0_dp)/dn
     IF(wj .GT. xx(jj+1)) jj = jj + 1
!
     sx(jj) = sx(jj) + (REAL(xj)**2)+(AIMAG(xj)**2)
     ix(jj) = ix(jj) + 1
!
     kk = 1
     DO k = 1, j
        xk = CMPLX(xr(k),xi(k))
        wk = (DBLE(k) - 1.0_dp)/dn
!
        IF(wk .GT. xx(kk+1)) kk = kk + 1
        IF((2*(j-1)+(k-1)) .GT. n) GO TO 10
!
        jpk = j+k-1
        B(jj,kk) = B(jj,kk) + xj*xk*CMPLX(xr(jpk),-xi(jpk))
!
        i1(jj,kk) = i1(jj,kk) + 1
        IF((j .EQ. k) .OR. ((2*(j-1)+(k-1)) .EQ. n)) THEN
           iq(jj,kk) = iq(jj,kk) + 2
        ELSE
           iq(jj,kk) = iq(jj,kk) + 1
        END IF
     END DO
10   CONTINUE
  END DO
!
  DO j = 1, m
     sx(j) = sx(j)/(DBLE(ix(j))*dn)
     DO k = 1, j
        IF(i1(j,k) .GT. 0) B(j,k) = B(j,k)/(dn*DBLE(i1(j,k)))
     END DO
  END DO
!
  DO j = 1, m
     sxj = sx(j)
     DO k = 1, m
        IF(iq(j,k) .EQ. 0) GO TO 20
        sxk = sx(k)
!
!  Now to estimate the autospectra at frequency that is the sum of the
!  horizontal and vertical frequency.  But this sum is a grid point.
!  Thus we use the average of the two surrounding values; that is,
!  fhat(omega(j)+omega(k)) = (sx(j+k-1)+sx(j+k))/2
!
        IF((j+k) .LE. m) THEN
           sx1 = sx(j+k)
        ELSE
           sx1 = sx(2*m+1-j-k)
        END IF
        IF((j+k-1) .LE. m) THEN
           sx2 = sx(j+k-1)
        ELSE
           sx2 = sx(2*m-j-k+2)
        END IF
        sxjpk  = (sx1 + sx2)/2.0_dp
        div    = (sxj*sxk*sxjpk*dn)/DBLE(i1(j,k))
!
        z(j,k) = 2.0_dp*(REAL(B(j,k))**2 + AIMAG(B(j,k))**2)/div
     END DO
20 END DO
!
  i1(1:m,1:m) = 0
  jj = 1
  DO j = 1, nq
     wj = (DBLE(j) - 1.0_dp)/dn
     IF(wj .GT. xx(jj+1)) jj = jj + 1
     kk = 1
     DO k = 1, nq
        wk = (DBLE(k) - 1.0_dp)/dn
        IF(wk .GT. xx(kk+1)) kk = kk + 1
        IF(((2*(j-1)+(k-1)) .GE. n) .OR. (k .GE. j)) THEN
           i1(jj,kk) = 0
           GO TO 30
        ELSE
           i1(jj,kk) = i1(jj,kk) + 1
        END IF
     END DO
30   CONTINUE
  END DO
!
  RETURN
END SUBROUTINE hinich_bispec
